package gov.usgs.earthquake.nshmp.netcdf.www;

import java.util.List;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfDataFilesGroundMotions;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfGroundMotions;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfVersion;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticData;
import gov.usgs.earthquake.nshmp.netcdf.www.Metadata.ServiceResponseMetadata;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfWsUtils.Key;
import gov.usgs.earthquake.nshmp.netcdf.www.Query.Service;
import gov.usgs.earthquake.nshmp.netcdf.www.Request.RequestData;
import gov.usgs.earthquake.nshmp.netcdf.www.Request.RequestDataSiteClass;
import gov.usgs.earthquake.nshmp.www.ResponseBody;
import gov.usgs.earthquake.nshmp.www.ResponseMetadata;
import gov.usgs.earthquake.nshmp.www.WsUtils;

import io.micronaut.http.HttpRequest;

/**
 * Handles service calls for ground motions.
 *
 * @see NetcdfController
 *
 * @author U.S. Geological Survey
 */
public class NetcdfServiceGroundMotions extends NetcdfService<Query> {

  static final String SERVICE_DESCRIPTION = "Get static ground motions from a NetCDF file";
  static final String X_LABEL = "Period (s)";
  static final String Y_LABEL = "Spectral Acceleration (g)";

  public NetcdfServiceGroundMotions(NetcdfDataFilesGroundMotions netcdfDataFiles) {
    super(netcdfDataFiles);
  }

  @Override
  ResponseBody<String, Metadata<Query>> getMetadataResponse(HttpRequest<?> request) {
    var metadata = new Metadata<Query>(request, this, SERVICE_DESCRIPTION);

    return ResponseBody.<String, Metadata<Query>> usage()
        .metadata(new ResponseMetadata(NetcdfVersion.appVersions()))
        .name(getServiceName())
        .request(NetcdfWsUtils.getRequestUrl(request))
        .response(metadata)
        .url(NetcdfWsUtils.getRequestUrl(request))
        .build();
  }

  @Override
  String getServiceName() {
    return String.format("AASHTO-%d Web Services", netcdfDataFiles().aashtoYear());
  }

  NetcdfGroundMotions netcdf(Location location) {
    return netcdfDataFiles().netcdf(location);
  }

  @Override
  NetcdfDataFilesGroundMotions netcdfDataFiles() {
    return (NetcdfDataFilesGroundMotions) netcdfDataFiles;
  }

  @Override
  String processRequest(HttpRequest<?> httpRequest, Query query, Service service) {
    var site = Location.create(query.longitude, query.latitude);
    var requestData = new RequestData(site, query.format);
    var url = NetcdfWsUtils.getRequestUrl(httpRequest);

    switch (service) {
      case CURVES:
        return toResponseFromList(processCurves(requestData, url));
      case CURVES_BY_SITE_CLASS:
        requestData = new RequestDataSiteClass(site, query.siteClass, query.format);
        return toResponse(processCurvesSiteClass(
            (RequestDataSiteClass) requestData, url));
      default:
        throw new RuntimeException("Netcdf service [" + service + "] not found");
    }
  }

  ResponseBody<RequestDataSiteClass, ResponseData<ServiceResponseMetadata>> processCurvesSiteClass(
      RequestDataSiteClass request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.longitude);
    WsUtils.checkValue(Key.SITE_CLASS, request.siteClass);
    var curves = netcdf(request.site).staticData(request.site, request.siteClass);
    var responseData = toResponseData(request, curves);

    return ResponseBody.<RequestDataSiteClass, ResponseData<ServiceResponseMetadata>> success()
        .metadata(new ResponseMetadata(NetcdfVersion.appVersions()))
        .name(getServiceName())
        .request(request)
        .response(responseData)
        .url(url)
        .build();
  }

  ResponseBody<RequestData, List<ResponseData<ServiceResponseMetadata>>> processCurves(
      RequestData request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.longitude);
    var curves = netcdf(request.site).staticData(request.site);
    var responseData = toList(request, curves);

    return ResponseBody.<RequestData, List<ResponseData<ServiceResponseMetadata>>> success()
        .metadata(new ResponseMetadata(NetcdfVersion.appVersions()))
        .name(getServiceName())
        .request(request)
        .response(responseData)
        .url(url)
        .build();
  }

  List<ResponseData<ServiceResponseMetadata>> toList(
      RequestData requestData,
      StaticData<XySequence> curves) {
    return curves.entrySet().stream()
        .map(entry -> {
          var request =
              new RequestDataSiteClass(requestData.site, entry.getKey(), requestData.format);
          return toResponseData(request, entry.getValue());
        })
        .collect(Collectors.toList());
  }

  ResponseData<ServiceResponseMetadata> toResponseData(
      RequestDataSiteClass request,
      XySequence curves) {
    var metadata = new ServiceResponseMetadata(
        request.siteClass,
        X_LABEL,
        Y_LABEL);
    return new ResponseData<>(metadata, curves);
  }
}
