package gov.usgs.earthquake.nshmp.netcdf.www;

import java.util.List;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfDataFilesHazardCurves;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfHazardCurves;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfVersion;
import gov.usgs.earthquake.nshmp.netcdf.Nshm;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticData;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticDataHazardCurves;
import gov.usgs.earthquake.nshmp.netcdf.www.HazardMetadata.HazardResponseMetadata;
import gov.usgs.earthquake.nshmp.netcdf.www.NetcdfWsUtils.Key;
import gov.usgs.earthquake.nshmp.netcdf.www.Query.Service;
import gov.usgs.earthquake.nshmp.netcdf.www.RequestHazardCurves.HazardRequestData;
import gov.usgs.earthquake.nshmp.netcdf.www.RequestHazardCurves.HazardRequestDataImt;
import gov.usgs.earthquake.nshmp.netcdf.www.RequestHazardCurves.HazardRequestDataSiteClass;
import gov.usgs.earthquake.nshmp.www.ResponseBody;
import gov.usgs.earthquake.nshmp.www.ResponseMetadata;
import gov.usgs.earthquake.nshmp.www.WsUtils;

import io.micronaut.http.HttpRequest;

/**
 * Handles service calls for hazard curves.
 *
 * @see NetcdfController
 *
 * @author U.S. Geological Survey
 */
public class NetcdfServiceHazardCurves extends NetcdfService<HazardQuery> {

  static final String SERVICE_DESCRIPTION = "Get static hazard curves from a NetCDF file";
  static final String X_LABEL = "Ground Motion (g)";
  static final String Y_LABEL = "Annual Frequency of Exceedence";

  public NetcdfServiceHazardCurves(NetcdfDataFilesHazardCurves netcdfDataFiles) {
    super(netcdfDataFiles);
  }

  @Override
  ResponseBody<String, Metadata<HazardQuery>> getMetadataResponse(HttpRequest<?> request) {
    var metadata = new Metadata<HazardQuery>(request, this, SERVICE_DESCRIPTION);
    return ResponseBody.<String, Metadata<HazardQuery>> usage()
        .metadata(new ResponseMetadata(NetcdfVersion.appVersions()))
        .name(getServiceName())
        .request(NetcdfWsUtils.getRequestUrl(request))
        .response(metadata)
        .url(NetcdfWsUtils.getRequestUrl(request))
        .build();
  }

  @Override
  public Service getService(HazardQuery query) {
    if (query.imt != null) {
      return Service.CURVES_BY_IMT;
    } else if (query.siteClass != null) {
      return Service.CURVES_BY_SITE_CLASS;
    } else {
      return Service.CURVES;
    }
  }

  @Override
  String getServiceName() {
    return "Static Hazard Curves for NSHMs";
  }

  @Override
  List<SourceModel> getSourceModels() {
    return netcdfDataFiles().stream()
        .map(HazardSourceModel::new)
        .collect(Collectors.toList());
  }

  /**
   * Returns the {@link NetcdfHazardCurves} associated with a given {@link Nshm
   * NSHM}.
   *
   * @param nshm The NSHM to get the NetCDF data
   */
  NetcdfHazardCurves netcdf(Nshm nshm) {
    return netcdfDataFiles().netcdf(nshm);
  }

  @Override
  NetcdfDataFilesHazardCurves netcdfDataFiles() {
    return (NetcdfDataFilesHazardCurves) netcdfDataFiles;
  }

  ResponseBody<HazardRequestData, List<List<ResponseData<HazardResponseMetadata>>>> processCurves(
      HazardRequestData request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.longitude);
    var curves = netcdf(request.nshm).staticData(request.site);
    var curvesAsList = toList(request, curves);

    return ResponseBody
        .<HazardRequestData, List<List<ResponseData<HazardResponseMetadata>>>> success()
        .metadata(new ResponseMetadata(NetcdfVersion.appVersions()))
        .name(getServiceName())
        .request(request)
        .response(curvesAsList)
        .url(url)
        .build();
  }

  ResponseBody<HazardRequestDataSiteClass, List<ResponseData<HazardResponseMetadata>>> processCurvesSiteClass(
      HazardRequestDataSiteClass request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.longitude);
    WsUtils.checkValue(Key.SITE_CLASS, request.siteClass);
    var curves = netcdf(request.nshm).staticData(request.site, request.siteClass);
    var curvesAsList = toList(request, curves);

    return ResponseBody
        .<HazardRequestDataSiteClass, List<ResponseData<HazardResponseMetadata>>> success()
        .metadata(new ResponseMetadata(NetcdfVersion.appVersions()))
        .name(getServiceName())
        .request(request)
        .response(curvesAsList)
        .url(url)
        .build();
  }

  ResponseBody<HazardRequestDataSiteClass, ResponseData<HazardResponseMetadata>> processCurvesImt(
      HazardRequestDataImt request,
      String url) {
    WsUtils.checkValue(Key.LATITUDE, request.latitude);
    WsUtils.checkValue(Key.LONGITUDE, request.longitude);
    WsUtils.checkValue(Key.SITE_CLASS, request.siteClass);
    WsUtils.checkValue(Key.IMT, request.imt);
    var curves = netcdf(request.nshm).staticData(request.site, request.siteClass);

    return ResponseBody.<HazardRequestDataSiteClass, ResponseData<HazardResponseMetadata>> success()
        .metadata(new ResponseMetadata(NetcdfVersion.appVersions()))
        .name(getServiceName())
        .request(request)
        .response(toResponseData(request, request.imt, curves.get(request.imt)))
        .url(url)
        .build();
  }

  @Override
  String processRequest(HttpRequest<?> httpRequest, HazardQuery query,
      Service service) {
    var site = Location.create(query.longitude, query.latitude);
    var url = NetcdfWsUtils.getRequestUrl(httpRequest);

    switch (service) {
      case CURVES:
        var requestData = new HazardRequestData(query.nshm, site, query.format);
        return toResponseFromListOfList(processCurves(requestData, url));
      case CURVES_BY_SITE_CLASS:
        var requestDataSiteClass =
            new HazardRequestDataSiteClass(query.nshm, site, query.siteClass, query.format);
        return toResponseFromList(processCurvesSiteClass(requestDataSiteClass, url));
      case CURVES_BY_IMT:
        var requestDataImt =
            new HazardRequestDataImt(query.nshm, site, query.siteClass, query.imt, query.format);
        return toResponse(processCurvesImt(requestDataImt, url));
      default:
        throw new RuntimeException("Netcdf service [" + service + "] not found");
    }
  }

  private String toResponseFromListOfList(
      ResponseBody<HazardRequestData, List<List<ResponseData<HazardResponseMetadata>>>> serviceResponse) {
    if (serviceResponse.getRequest().format == ResponseFormat.CSV) {
      var csvResponse =
          toCsvFromListOfList(serviceResponse.getRequest(), serviceResponse.getResponse());
      return String.format("%s\n\n%s", getServiceName(), csvResponse);
    } else {
      return NetcdfWsUtils.GSON.toJson(serviceResponse);
    }
  }

  private String toCsvFromListOfList(HazardRequestData requestData,
      List<List<ResponseData<HazardResponseMetadata>>> responseData) {
    return responseData.stream().map(responses -> toCsvResponseFromList(requestData, responses))
        .collect(Collectors.joining("\n\n"));
  }

  List<ResponseData<HazardResponseMetadata>> toList(
      HazardRequestDataSiteClass request,
      StaticDataHazardCurves curves) {
    return curves.entrySet().stream()
        .map((entry) -> toResponseData(request, entry.getKey(), entry.getValue()))
        .collect(Collectors.toList());
  }

  List<List<ResponseData<HazardResponseMetadata>>> toList(
      HazardRequestData requestData,
      StaticData<StaticDataHazardCurves> curves) {
    return curves.entrySet().stream()
        .map(entry -> {
          var request =
              new HazardRequestDataSiteClass(requestData.nshm, requestData.site, entry.getKey(),
                  requestData.format);
          return toList(request, entry.getValue());
        })
        .collect(Collectors.toList());
  }

  ResponseData<HazardResponseMetadata> toResponseData(
      HazardRequestDataSiteClass request,
      Imt imt,
      XySequence curves) {
    var metadata =
        new HazardResponseMetadata(request.nshm, request.siteClass, imt, X_LABEL, Y_LABEL);
    return new ResponseData<>(metadata, curves);
  }

  static class HazardSourceModel extends SourceModel {
    public final Nshm nshm;
    public final List<Imt> imts;

    HazardSourceModel(NetcdfHazardCurves netcdf) {
      super(netcdf);
      nshm = netcdf.nshm();
      imts = netcdf.netcdfData().imts().stream().sorted().collect(Collectors.toList());
    }
  }
}
