package gov.usgs.earthquake.nshmp.netcdf;

import java.nio.file.Files;
import java.nio.file.Path;
import java.util.logging.Level;
import java.util.logging.Logger;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.data.BoundingData;
import gov.usgs.earthquake.nshmp.netcdf.data.NetcdfData;
import gov.usgs.earthquake.nshmp.netcdf.data.NetcdfShape;
import gov.usgs.earthquake.nshmp.netcdf.data.StaticData;
import gov.usgs.earthquake.nshmp.netcdf.reader.Reader;

/**
 * Abstarct class for NetCDF types: hazard curves and ground motions.
 *
 * @author U.S. Geological Survey
 */
public abstract class Netcdf<T> implements Comparable<Netcdf<T>> {

  protected final Path netcdfPath;
  protected final NetcdfData netcdfData;
  protected NetcdfShape netcdfShape;

  private static final Logger LOGGER = Logger.getLogger("ucar");

  static {
    /* Update ucar logger */
    LOGGER.setLevel(Level.SEVERE);
  }

  public Netcdf(Path netcdfPath) {
    this.netcdfPath = netcdfPath;

    if (Files.notExists(netcdfPath)) {
      throw new IllegalArgumentException("Path to Netcdf file [" + netcdfPath + "] does not exist");
    }

    var reader = getNetcdfData(netcdfPath);
    netcdfData = reader.netcdfData();
    netcdfShape = reader.netcdfShape();
  }

  /**
   * Returns the bounding data from a specific site.
   *
   * @param site The site to get the bounding data
   */
  public abstract BoundingData<T> boundingData(Location site);

  /**
   * Returns the NetCDF data.
   */
  public abstract NetcdfData netcdfData();

  /**
   * Returns the NetCDF path.
   */
  public Path netcdfPath() {
    return netcdfPath;
  }

  /**
   * Returns the NetCDF shape`
   */
  public NetcdfShape netcdfShape() {
    return netcdfShape;
  }

  /**
   * Returns the static data from a specific site.
   *
   * @param site The site to get the static data
   */
  public abstract StaticData<T> staticData(Location site);

  /**
   * Returns the static data from a site and site class.
   *
   * @param site The site to get the static data
   * @param siteClass The site class to get the static data
   */
  public abstract T staticData(Location site, NehrpSiteClass siteClass);

  @Override
  public int compareTo(Netcdf<T> that) {
    return this.netcdfData().scienceBaseMetadata().label
        .compareTo(that.netcdfData().scienceBaseMetadata().label);
  }

  abstract Reader getNetcdfData(Path netcdfPath);
}
